"""Tests for `gconfigs.backends` package.
"""
import os

import pytest

from gconfigs.backends import LocalEnv, LocalMountFile, DotEnv, RootDirectoryNotFound


def test_local_env():
    """Tests for `gconfigs.backends.LocalEnv`
    """
    os.environ.update({"GCONFIGS_ENV_TEST": "env-test-1"})
    backend = LocalEnv()

    assert "GCONFIGS_ENV_TEST" in backend.keys()
    assert backend.get("GCONFIGS_ENV_TEST") == "env-test-1"
    with pytest.raises(KeyError):
        backend.get("GCONFIGS_NON-EXISTENT-ENV-KEY")


def test_local_mount_file_generic():
    """Tests for `gconfigs.backends.LocalMountFile` with root_dir='./tests/files/configs'
    """
    with pytest.raises(RootDirectoryNotFound):
        backend = LocalMountFile(root_dir="./NON-EXISTENT-DIR")
        assert backend.root_dir

    backend = LocalMountFile(root_dir="./tests/files/configs")
    assert "CONFIG_TEST" in backend.keys()
    assert backend.get("CONFIG_TEST") == "config-test"

    with pytest.raises(FileNotFoundError):
        backend.get("NON-EXISTENT-CONFIG")


def test_dotenv():
    backend = DotEnv()
    with pytest.raises(Exception, match=r".*you didn't loaded the your dotenv file.*"):
        backend.get("CONFIG-1")

    backend.load_file("./tests/files/config-files/.env")
    assert "CONFIG-1" in backend.keys()
    with pytest.raises(KeyError):
        backend.get("NON-EXISTENT-CONFIG-KEY")

    # `.get` must filter break lines
    assert "\n" not in backend.get("CONFIG-1") and "\r" not in backend.get(
        "CONFIG-1"
    ), "Breaklines on values must be removed."

    assert backend.get("CONFIG-1") == "config-1"
    assert "config-key-with-spaces" in backend.keys(), "Keys must be stripped."
    assert backend.get(
        "config-value-with-spaces"
    ) == " config value with spaces", "values should NOT be stripped."

    assert "COMMENTED-CONFIG" not in backend.keys()
    assert backend.get("CONFIG-EMPTY-VALUE") == "", "Empty values must be empty string."

    # just seeing if we can iterate over all configs
    for key in backend.keys():
        assert not key.startswith(("#", ";", "[")), "Invalid lines must be removed."
        assert "=" not in key
        # the `CONFIG-EMPTY-VALUE` will be an empty value, so, no good to assert
        if key == "CONFIG-EMPTY-VALUE":
            continue

        assert backend.get(key)

    # Empty .env file
    with pytest.raises(Exception, match=r".*You have no configs to look for.*"):
        backend.load_file("./tests/files/config-files/.env-empty")

    with pytest.raises(FileNotFoundError):
        backend.load_file("./tests/files/config-files/NON-EXISTENT-DOTENV-FILE")



"""Tests for `gconfigs` package.
- There are some asserts that may seem obvious, that's just because they
act like a restriction to things I think, at the moment, that we may have to
talk before changing the way it's implemented.
"""

from gconfigs import GConfigs
from . import DummyBackend
import gconfigs

import pytest

import json


def test_bad_instatiation():

    class MissingGet:
        pass

    class MissingKeys:
        pass

    with pytest.raises(AttributeError):
        GConfigs(backend=MissingGet)
    with pytest.raises(AttributeError):
        GConfigs(backend=MissingKeys)

    with pytest.raises(TypeError, match=r".*required positional argument: 'backend'.*"):
        GConfigs()


def test_basics():
    # main endpoints
    gconfigs.envs
    gconfigs.dotenvs
    gconfigs.configs
    gconfigs.secrets
    # basic expected
    configs_ = GConfigs(backend=DummyBackend)
    configs_.strip
    configs_.object_type_name
    configs_.get
    configs_.__call__
    # iterator
    configs_.iterator
    configs_.__next__
    configs_.__iter__
    # fancy xD
    configs_.json
    configs_.__contains__
    configs_.__len__
    configs_.__repr__

    assert configs_.object_type_name == "KeyValue", "'object_type_name' default value should be 'KeyValue'"
    assert configs_.strip, "'strip' default value should be boolean 'True'"

    configs = GConfigs(backend=DummyBackend, object_type_name="DummyConfig")
    assert configs.object_type_name == "DummyConfig"
    configs.object_type_name = "DummyConfigChanged"
    assert configs.object_type_name == "DummyConfigChanged"


def test_get_configs_info():
    configs = GConfigs(backend=DummyBackend)
    # if given key is in configs (backend.keys())
    assert "CONFIG-1" in configs
    assert "NON-EXISTENT-CONFIG" not in configs
    assert len(configs) == len(configs._backend.data)
    assert configs("CONFIG-1") == "config-1"

    # Non existent config
    with pytest.raises(KeyError):
        configs("NON-EXISTENT-CONFIG")

    # Non existent config - return default instead
    assert configs("NON-EXISTENT-CONFIG", default="default") == "default"
    # default argument can be anything
    # achieved by using the `gconfigs.gconfigs.NOTSET` trick
    assert configs("NON-EXISTENT-CONFIG", default=None) is None
    assert configs("NON-EXISTENT-CONFIG", default=True) is True
    assert configs("NON-EXISTENT-CONFIG", default=False) is False


def test_get_configs_strip_value_or_not():
    configs = GConfigs(backend=DummyBackend)
    assert configs.strip
    # strip default: True
    assert configs(" white space key ") == "white space value" and configs(
        "space-only-value"
    ) == "", "Returning value should be stripped."
    # override default
    assert configs(
        " white space key ", strip=False
    ) == " white space value " and configs(
        "space-only-value", strip=False
    ) == "  ", "Returning value should ALLOW blank spaces."

    # no strip
    configs2 = GConfigs(backend=DummyBackend, strip=False)
    assert configs2.strip is False
    # strip default: False
    assert configs2(" white space key ") == " white space value " and configs2(
        "space-only-value"
    ) == "  ", "Returning value should ALLOW blank spaces."
    # override default
    assert configs2(
        " white space key ", strip=True
    ) == "white space value" and configs2(
        "space-only-value", strip=True
    ) == "", "Returning value should be stripped."
    configs2.strip = True
    assert configs2.strip


def test_get_typed_configs():
    """Check if gConfigs deal with typed configs
    Most backends will probably return strings as value of a config, but
    if someone cast the value it's nice to know if gConfigs is ok with this.
    """
    configs = GConfigs(backend=DummyBackend)
    assert configs("CONFIG-1") == "config-1"
    assert configs("CONFIG-NONE") is None
    assert configs("CONFIG-TRUE") is True
    assert configs("CONFIG-FALSE") is False
    assert configs("CONFIG-INT") == 1
    assert configs("CONFIG-FLOAT") == 1.1
    assert configs("CONFIG-LIST") == [1, 1.1, "a"]
    assert configs("CONFIG-TUPLE") == (1, 1.1, "a")
    assert configs("CONFIG-DICT") == {"a": 1, "b": "b"}


def test_get_and_cast_value():
    configs = GConfigs(backend=DummyBackend)
    # BOOLEAN
    assert configs.as_bool("CONFIG-TRUE")  # when backend return a bool already
    assert configs.as_bool("CONFIG-TRUE-STRING")
    assert isinstance(configs.as_bool("CONFIG-TRUE-STRING"), bool)
    assert not configs.as_bool("CONFIG-FALSE-STRING")
    assert isinstance(configs.as_bool("CONFIG-FALSE-STRING"), bool)
    assert configs.as_bool("NON-EXISTENT-CONFIG", default=True)
    assert isinstance(configs.as_bool("NON-EXISTENT-CONFIG", default=True), bool)
    with pytest.raises(ValueError, match=r".*Could not cast the value.*"):
        configs.as_bool("CONFIG-NONE")

    # LIST
    assert configs.as_list("CONFIG-LIST") == [1, 1.1, "a"]
    assert configs.as_list("CONFIG-TUPLE") == [1, 1.1, "a"]
    assert configs.as_list("CONFIG-LIST-STRING-JSON-STYLE") == [1, 1.1, "a"]
    with pytest.raises(ValueError, match=r".*Could not cast the value.*"):
        configs.as_list("CONFIG-NONE")

    # DICT
    assert configs.as_dict("CONFIG-DICT") == {"a": 1, "b": "b"}
    assert configs.as_dict("CONFIG-DICT-JSON-STYLE") == {"a": 1, "b": "b"}
    with pytest.raises(ValueError, match=r".*Could not cast the value.*"):
        configs.as_dict("CONFIG-NONE")

    # trigger exception on GConfigs._cast()
    with pytest.raises(ValueError, match=r".*must be a valid json value.*"):
        configs.as_list("CONFIG-LIST-STRING-JSON-STYLE-BROKEN")


def test_iterator():
    configs = GConfigs(backend=DummyBackend)
    # iterate with forloop
    for config in configs:
        # Make sure the object is a namedtuple
        assert issubclass(
            config.__class__, tuple
        ) and config._fields, "Looks like the iterator is not returning a namedtuple anymore."
        assert config
        assert config.key
        # guarantee the namedtuple is being properly created
        assert configs(
            config.key
        ) == config.value, "Looks like the namedtuple is not being properly created."

    # after an iterate in forloops it should have nothing left for next()
    with pytest.raises(StopIteration):
        next(configs)

    # but it's possible to iterator as many as you can with `self.iterator()`
    iter_configs = configs.iterator()
    for config in iter_configs:
        assert config


def test_json():
    configs = GConfigs(backend=DummyBackend)
    json_ = configs.json()
    assert json_
    # let's turn the json_ into a dict so we can see if it preserved all the configs
    configs_data = json.loads(json_)
    assert all(key in configs._backend.data for key in configs_data)

    assert next(
        configs
    ), "If `configs` is not iterable after `configs.json()` it's because it's not using the iterator properly."



"""
Backends for gConfigs
~~~~~~~~~~~~~~~~~~~~~

Backends are just simple classes with at least two methods implemented:
`get` and `keys`.

Example:

    class RedisBackend:
        def keys(self):
            # return a iterable of all keys available
            return available_keys

        def get(self, key: str):
            # this method receive a key (identifier of a config)
            # and return its respective value
            return value

Notes:
    - If it's not possible to provide a `.keys` method, just declare with:
    raise NotImplementedError. Of course it will limit the goodies of gConfigs.
    - For errors on `.get` method just throw exceptions.
    (Config doesn't exists, you don't have permission, stuff like that)
    See `GConfigs.get` and you'll see that it has a `default` parameter,
    and of course if you provide a default value it will not throw a exception.
"""
from pathlib import Path
import os


class LocalEnv:

    def keys(self):
        return os.environ.keys()

    def get(self, key, **kwargs):
        value = os.environ.get(key)
        if value is None:
            raise KeyError(
                f"Environment variable '{key}' not set. Check for any "
                "misconfiguration or misspelling of the variable name."
            )

        return value


class LocalMountFile:

    def __init__(self, root_dir="/", pattern="*"):
        self.pattern = pattern
        self.root_dir = root_dir

    @property
    def root_dir(self):
        if not self._root_dir.exists():
            raise RootDirectoryNotFound(
                f"The root directory {self._root_dir} doesn't exist."
            )

        return self._root_dir

    @root_dir.setter
    def root_dir(self, root_dir):
        self._root_dir = Path(root_dir)

    def keys(self):
        for item in self.root_dir.glob(self.pattern):
            if item.is_file():
                yield item.name

    def get(self, key, **kwargs):
        file = self.root_dir / key
        if file.exists():
            return file.read_text()

        raise FileNotFoundError(
            f"Check if your files are mounted on {self.root_dir}. "
            "And remember to check if your system is case sensitive."
        )


class DotEnv:

    def keys(self):
        return self._data.keys()

    def get(self, key, **kwargs):
        if not hasattr(self, "_dotenv_file"):
            raise Exception("It seems like you didn't loaded the your dotenv file yet.")

        value = self._data.get(key)
        if value is None:
            raise KeyError(
                f"The config '{key}' is not set on {self._dotenv_file}. Check "
                "for any misconfiguration or misspelling of the variable name."
            )

        return value

    def load_file(self, filepath):
        self._dotenv_file = filepath
        self._data = {}
        with open(self._dotenv_file) as file:
            for line in file.readlines():
                # ignore comments, section title or invalid lines
                if line.startswith(("#", ";", "[")) or "=" not in line:
                    continue

                # split on the first =, allows for subsequent `=` in strings
                key, value = line.split("=", 1)
                key = key.strip()

                if not (key and value):
                    continue

                self._data[key] = value.rstrip("\r\n")

        if not self._data:
            raise Exception(
                f"You have no configs to look for. You need to load your {self._dotenv_file} file."
            )


class RootDirectoryNotFound(FileNotFoundError):
    pass
